//
//  EMTI_Population.cpp
//  EMTI
//
//  Created by Anders Nikolai Fuglseth, 2015.

#include "EMTI_Population.hpp"

EMTI_Population::EMTI_Population (vector<int>* fullAttributeScope, vector<vector<unsigned>>* trainingSet, int maxTreeSize, int populationSize, int treeSize_fitnessWeight, float mutationRate, float crossOverRateParameter, bool rejectWorseGeneration, int parentSelectionType, int adultSelectionType, bool verbose, bool verboseplus, bool writePlotData, std::string plotOutputLocation, int tournamentSizeParameter, float epsilonParameter, float spiceRatio) {
    
    /* initialize random seed based on current time: */
    //srand ((int)time(NULL));
    
    /* store parameters: */
    if (populationSize < 2) this->populationSize = 2;
    else this->populationSize = populationSize;
    this->treeSize_fitnessWeight = treeSize_fitnessWeight;
    this->mutationRateParameter = mutationRate;
    this->crossOverRateParameter = crossOverRateParameter;
    this->rejectWorseGeneration = rejectWorseGeneration;
    this->parentSelectionType = parentSelectionType;
    this->adultSelectionType = adultSelectionType;
    
    this->verboseplus = verboseplus;
    if (verboseplus) {this->verbose = true;}
    else this->verbose = verbose;
    
    this->writePlotData = writePlotData;
    this->plotOutputLocation = plotOutputLocation;
    
    this->tournamentSizeParameter = tournamentSizeParameter;
    this->epsilonParameter = epsilonParameter;
    
    this->spiceRatio = spiceRatio;
    
    this->trainingSet = trainingSet;
    
    this->fullAttributeScope = new vector<int>(*fullAttributeScope);
    
    this->maxTreeSize = maxTreeSize;
    if (maxTreeSize > 0) maintainMaxTreeSize = true;
    
    /* create new random population: */
    children = new vector<EMTI_TreeCreature*>;
    adults = new vector<EMTI_TreeCreature*>;
    
    float currentBestFitness = std::numeric_limits<float>::min();;
    for (int i = 0; i<populationSize; i++) {
        
        EMTI_TreeCreature* ind = new EMTI_TreeCreature(fullAttributeScope);
        ind->createRandomTree();
        ind->assignFitness(trainingSet, treeSize_fitnessWeight);
        if (ind->getFitness() > currentBestFitness) {
            currentBestFitness = ind->getFitness();
            childrenBestIndividual = ind;
        }
        children->push_back(ind);
    }
    adultBestIndividual = childrenBestIndividual;
    
    //subSetSize = (int)trainingSet->size() / 2;
}

EMTI_Population::~EMTI_Population () {
    for (EMTI_TreeCreature* individual : *adults) {
        delete individual;
    }
    adults->clear();
    delete adults;
    
    for (EMTI_TreeCreature* individual : *children) {
        delete individual;
    }
    children->clear();
    delete children;
    
    delete fullAttributeScope;
}

void EMTI_Population::do_evolution_cycle () {
    
    if (rejectWorseGeneration && childrenBestIndividual->getFitness() < adultBestIndividual->getFitness()) {
        childrenBestIndividual = adultBestIndividual;
        for (EMTI_TreeCreature* child : *children) {
            delete child;
        }
        children->clear();
        delete children;
    }
    
    else {
        
        
        /*perform generation shift*/
        if (adultSelectionType == EMTI_AdultSelection::full_generational_replacement_selection) {
            adults = EMTI_AdultSelection::full_generational_replacement(children, adults);
            adultBestIndividual = childrenBestIndividual;
        }
        else if (adultSelectionType == EMTI_AdultSelection::overproduction_selection) {
            adults = EMTI_AdultSelection::overproduction(children, adults, populationSize);
            adultBestIndividual = childrenBestIndividual;
        }
        else if (adultSelectionType == EMTI_AdultSelection::generational_mixing_selection) {
            adults = EMTI_AdultSelection::generational_mixing(children, adults);
            adultBestIndividual = adults->front();
        }
        else if (adultSelectionType == EMTI_AdultSelection::generational_mixing_add_spice_selection) {
            adults = EMTI_AdultSelection::generational_mixing_add_spice(children, adults, spiceRatio, fullAttributeScope, trainingSet, treeSize_fitnessWeight);
            adultBestIndividual = adults->front();
        }
        
        
    }
    
    //int childrenSize = (int)adults->size();
    int childrenSize = populationSize;
    if (adultSelectionType == 1) { //overproduction
        childrenSize = childrenSize * 2;
    }
    
    /*create new children*/
    children = new vector<EMTI_TreeCreature*>;
    
    while (children->size() < childrenSize) {
        
        pair<EMTI_TreeCreature*, EMTI_TreeCreature*> parents;
        
        if (parentSelectionType == EMTI_ParentSelection::fitness_proportionate_selection) {
            parents = EMTI_ParentSelection::fitness_proportionate(adults);
        }
        else if (parentSelectionType == EMTI_ParentSelection::tournament_selection) {
            parents = EMTI_ParentSelection::tournament(adults, tournamentSizeParameter, epsilonParameter);
        }
        else if (parentSelectionType == EMTI_ParentSelection::sigma_scaling_selection) {
            parents = EMTI_ParentSelection::sigma_scaling(adults, averageFitness, sdFitness);
        }
        
        pair<EMTI_TreeCreature*, EMTI_TreeCreature*> childpair = EMTI_TreeCreature::crossover(parents, crossOverRateParameter, fullAttributeScope);
        
        
        mutationsInCycleCount += childpair.first->mutate(mutationRateParameter);
        mutationsInCycleCount += childpair.second->mutate(mutationRateParameter);
        
        if (maintainMaxTreeSize && childpair.first->tree->nodesCount > maxTreeSize) {
            delete childpair.first;
            childpair.first = new EMTI_TreeCreature(fullAttributeScope);
            childpair.first->createRandomTree();
        }
        if (maintainMaxTreeSize && childpair.second->tree->nodesCount > maxTreeSize) {
            delete childpair.second;
            childpair.second = new EMTI_TreeCreature(fullAttributeScope);
            childpair.second->createRandomTree();
        }
        
        children->push_back(childpair.first);
        children->push_back(childpair.second);
        
        
    }
    
    /*determine fitness for all children and find best fitness*/
    float currentBestFitness = std::numeric_limits<float>::min();
    
    //    vector<vector<unsigned>>* subSet = getTrainingSetPart(trainingSet);
    
    for (EMTI_TreeCreature* child : *children) {
        
        
        child->assignFitness(trainingSet, treeSize_fitnessWeight);
        
        if (child->getFitness() > currentBestFitness) {
            currentBestFitness = child->getFitness();
            childrenBestIndividual = child;
        }
    }
    
    //    currentBestFitness = std::numeric_limits<float>::min();
    //    for (EMTI_TreeCreature* adult : *adults) {
    //        adult->assignFitness(subSet, treeSize_fitnessWeight);
    //        if (adult->getFitness() > currentBestFitness) {
    //            currentBestFitness = adult->getFitness();
    //            adultBestIndividual = adult;
    //        }
    //    }
    //
    //    delete subSet;
    //
    
    /*write plot data*/
    giveGenerationInfoOutput(verbose, verboseplus, writePlotData);
    
    cycleCount++;
    
}


void EMTI_Population::doPostCycleEndNodeOptimizations(int k, bool verbose) {
    
    int changeCount = 0;
    
    if (adultBestIndividual != NULL) {
        
        adultBestIndividual-> determineEndNodeAccuraciesAndTreeTraffic (trainingSet);
        
        for (EMTI_TreeIterator i = adultBestIndividual->tree->begin(); i != adultBestIndividual->tree->end(); ++i) {
            if (i.node->traffic == 0 && i.node->isEndNode) { //bugfix 2016.04.20
                //if (i.node->traffic == 0 && i.node->parent != NULL) { //old version
                
                if (verbose) cout << "postOptimizations: [" << i.node << "] " << i.node->value;
                
                
                EMTI_TreeNode* node = i.node;
                map<int,int> entryMap;
                
                while (!node->isRoot) {
                    entryMap[node->parent->value] = node->childIndex;
                    
                    node = node->parent;
                }
                vector<unsigned>* entry = new vector<unsigned>();
                int n = 0;
                while (entry->size() < fullAttributeScope->size()-1) {
                    if (entryMap.find(n) != entryMap.end()) {
                        entry->push_back(entryMap[n]);
                    }
                    else {
                        entry->push_back((*fullAttributeScope)[n]); //setting a value that doesn't exist in the domain
                    }
                    n++;
                }
                EMTI_KNearestClassifier kN(trainingSet);
                
                int category = kN.classify(entry, /*k*/k, /*excludeLowerSimilarityMeasures*/false, /*verbose*/false);
                
                i.node->value = category;
                
                if (verbose) cout << " -> " << i.node->value << "\n";
                
                changeCount++;
                
                delete entry;
            }
        }
    }
    if (verbose) cout << "postOptimizations: looked at " << changeCount << " end nodes in total.\n";
}

vector<vector<unsigned>>* EMTI_Population::getTrainingSetPart (vector<vector<unsigned>>* trainingSet) {
    
    vector<vector<unsigned>>* subSet = new vector<vector<unsigned>>();
    
    if (currentPart + subSetSize >= trainingSet->size()) {
        long leftover = currentPart + subSetSize - trainingSet->size();
        subSet->insert(subSet->begin(), trainingSet->begin() + currentPart, trainingSet->end());
        subSet->insert(subSet->begin(), trainingSet->begin(), trainingSet->begin() + leftover);
        
        currentPart = (int)leftover;
        
    }
    else {
        subSet->insert(subSet->begin(), trainingSet->begin() + currentPart, trainingSet->begin() + currentPart + subSetSize);
        currentPart += subSetSize;
        
    }
    
    //cout << subSet->size() << "\n";
    
    return subSet;
}


int EMTI_Population::random_number(int start, int end) {
    int range = end - start + 1;
    int number;
    
    if (end>start) {
        number = rand() % range + start;
    }
    else {
        number = start;
    }
    
    return number;
}

double EMTI_Population::random_double(double start, double end) {
    double f = (double)rand() / RAND_MAX;
    return start + f * (end - start);
}


bool EMTI_Population::compareFitness (EMTI_TreeCreature* a, EMTI_TreeCreature* b) {
    return a->getFitness() > b->getFitness();
}

float EMTI_Population::measurePopulationDiversity (float sdSize, float sdAccu) {
    
    //    unordered_set<int> sizesFound;
    //    unordered_set<float> fitnessFound;
    //
    //    for (EMTI_TreeCreature* child : *children) {
    //        sizesFound.insert(child->tree->nodesCount);
    //        fitnessFound.insert(child->getFitness());
    //    }
    //    return sizesFound.size() + fitnessFound.size();
    return 0;
}


void EMTI_Population::giveGenerationInfoOutput (bool verbose, bool verboseplus, bool writePlotData) {
    
    averageFitness = 0;
    sdFitness = 0;
    
    averageSize = 0;
    sdSize = 0;
    
    double averageAccuracy = 0;
    double sdAccuracy = 0;
    
    for (EMTI_TreeCreature* child : *adults) {
        averageFitness += child->getFitness();
        averageSize += child->tree->nodesCount;
        averageAccuracy += child->classificationAccuracy;
    }
    
    averageFitness = averageFitness/adults->size();
    averageSize = averageSize/adults->size();
    averageAccuracy = averageAccuracy/adults->size();
    
    for (EMTI_TreeCreature* child : *adults) {
        sdFitness += pow(child->getFitness() - averageFitness, 2);
        sdSize += pow(child->tree->nodesCount - averageSize, 2);
        sdAccuracy += pow(child->classificationAccuracy - averageAccuracy, 2);
    }
    
    sdFitness = sqrt(sdFitness/adults->size());
    sdSize = sqrt(sdSize/adults->size());
    sdAccuracy = sqrt(sdAccuracy/adults->size());
    
    if (sdFitness < 0.001) sdFitness = 0.001;
    if (sdSize<0.001) sdSize=0.001;
    if (sdAccuracy<0.001) sdAccuracy=0.001;
    
    if (verbose || writePlotData || setCurrentGenerationInfoOutput) {
        
        diversity = measurePopulationDiversity(sdSize, sdAccuracy);
        
        if (verbose) {
            cout << "\n";
            cout << "Generation number " << cycleCount+1 << "\n";
            cout << "Average fitness=" << averageFitness << " SD=" << sdFitness << "\n";
            cout << "Average tree size=" << averageSize << " SD=" << sdSize << "\n";
            cout << "Average classification accuracy=" << averageAccuracy << "/" << trainingSet->size() << " SD=" << sdAccuracy << "\n";
            cout << "Mutations performed=" << mutationsInCycleCount << "\n";
            if (adultBestIndividual != NULL && adultBestIndividual->tree->root != NULL) {
                cout << "Best phenotype: [" << adultBestIndividual << "] " << adultBestIndividual->to_string() << "\n";
            }
            else {
                cout << "Best phenotype: [NULL] " << "\n";
            }
            if (verboseplus) {
                std::cout << "Phenotype population:\n";
                for (int i = 0; i < adults->size(); i++) {
                    std::cout << i << " [" << (*adults)[i] << "]: " << (*adults)[i]->to_string() << "\n";
                }
            }
        }
        if (writePlotData) {
            if (adultBestIndividual != NULL && adultBestIndividual->tree != NULL) {
                std::ofstream outfile;
                outfile.open(plotOutputLocation, std::ios_base::app);
                outfile << cycleCount << " ";;
                outfile << adultBestIndividual->getFitness();
                outfile << " " << averageFitness << " " << sdFitness
                << " " << diversity
                << " " << adultBestIndividual->tree->nodesCount << " " << averageSize << " " << sdSize
                << " " << adultBestIndividual->classificationAccuracy << " " << averageAccuracy << " " << sdAccuracy << "\n";
                outfile.close();
            }
        }
        if (setCurrentGenerationInfoOutput) {
            if (adultBestIndividual != NULL && adultBestIndividual->tree != NULL && adultBestIndividual->tree->root != NULL) {
                
                std::stringstream outputStream;
                
                outputStream << "Generation number " << cycleCount+1 << "\n";
                outputStream << "Average fitness=" << averageFitness << " SD=" << sdFitness << "\n";
                outputStream << "Average tree size=" << averageSize << " SD=" << sdSize << "\n";
                outputStream << "Average classification accuracy=" << averageAccuracy << "/" << trainingSet->size() << " SD=" << sdAccuracy << "\n";
                outputStream << "Mutations performed=" << mutationsInCycleCount << "\n";
                
                outputStream << "Best phenotype: [" << adultBestIndividual << "]\n" << adultBestIndividual->to_string() << "\n";
                
                currentGenerationInfoOutput = outputStream.str();
            }
            
            
        }
    }
    
    mutationsInCycleCount = 0;
}
