//
//  EMTI_DecisionTree.cpp
//  EMTI
//
//  Created by Anders Nikolai Fuglseth on 30.10.2015.

#include "EMTI_DecisionTree.hpp"



std::pair<int, int> operator +(const std::pair<int, int>& x, const std::pair<int, int>& y) {
    return std::make_pair(x.first + y.first, x.second + y.second);
}

EMTI_TreeNode::EMTI_TreeNode(EMTI_TreeNode* parent, int depth, int categoryScope, int value, int childIndex, bool isEndNode) {
    this->parent = parent;
    this->depth = depth;
    this->value = value;
    this->childIndex = childIndex;
    this->isEndNode = isEndNode;
    
    endNodeAccuracies = vector<int>(categoryScope, 0);
}

EMTI_TreeNode::~EMTI_TreeNode() {
    if (!isEndNode) {
        
        for ( auto &child : this->children ) {
            delete child;
        }
        
    }
}

EMTI_TreeNode::EMTI_TreeNode(const EMTI_TreeNode &b) {
    
    this->value = b.value;
    this->depth = b.depth;
    this->traffic = b.traffic;
    this->endNodeAccuracies = b.endNodeAccuracies;
    this->isEndNode = b.isEndNode;
    
    for (auto &child : b.children) {
        EMTI_TreeNode* newChild = new EMTI_TreeNode (*child);
        newChild->childIndex = child->childIndex;
        newChild->parent = this;
        this->children.push_back(newChild);
    }
}

void EMTI_TreeNode::setDepth(int depth) {
    
    this->depth = depth;
    
    for (auto &child: children) {
        child->setDepth(depth+1);
    }
}

void EMTI_TreeNode::clearEndNodeAccuracies() {
    fill(endNodeAccuracies.begin(), endNodeAccuracies.end(), 0);
}

EMTI_TreeIterator::EMTI_TreeIterator(EMTI_TreeNode* node) {
    this->node = node;
}

EMTI_TreeIterator EMTI_TreeIterator::operator++() {
    if (node->children.size() > 0) {
        node = node->children[0];
        return node;
    }
    else {
        EMTI_TreeNode* n =  node;
        while (n->parent != NULL && n->childIndex >= n->parent->children.size()-1) {
            n = n->parent;
        }
        if (n->parent == NULL) {
            node = nullptr;
            return node;
        }
        
        node = n->parent->children[n->childIndex+1];
        return n;
    }
    
}

bool EMTI_TreeIterator::operator!=(const EMTI_TreeIterator &rhs) const{
    if (rhs.node == NULL && this->node == NULL) return false;
    if (rhs.node == NULL) return true;
    if (this->node == NULL) return true;
    if (this->node == rhs.node) return false;
    return true;
}


EMTI_DecisionTree::EMTI_DecisionTree(int rootVal, int childrenCount, int categoryScope) {
    
    root = new EMTI_TreeNode(nullptr, 0 ,categoryScope, rootVal);
    
    this->categoryScope = categoryScope;
    
    int childIndex = 0;
    for (int i = 0; i < childrenCount; i++) {
        EMTI_TreeNode* leaf = new EMTI_TreeNode(root, root->depth+1, categoryScope, random_number(0, categoryScope-1), childIndex, true);
        root->children.push_back(leaf);
        childIndex++;
        
    }
    updateNodeCountAndList();
}

EMTI_DecisionTree::EMTI_DecisionTree(int categoryScope) {
    this->categoryScope = categoryScope;
}

EMTI_DecisionTree::~EMTI_DecisionTree() {
    
    delete root;
    
}

EMTI_DecisionTree::EMTI_DecisionTree (const EMTI_DecisionTree &b) {
    
    this->root = new EMTI_TreeNode(*b.root);
    this->categoryScope = b.categoryScope;
    this->nodesCount = b.nodesCount;
    this->endNodesCount = b.endNodesCount;
    this->decisionNodesCount = b.endNodesCount;
    
    updateNodeCountAndList();
}

int EMTI_DecisionTree::decide(vector<unsigned> instance) {
    
    EMTI_TreeNode* n = getEndNode(instance);
    
    if (n==NULL) return -1;
    else return getEndNode(instance)->value;
}

EMTI_TreeNode* EMTI_DecisionTree::getEndNode(vector<unsigned> instance, EMTI_TreeNode* node) {
    
    if (instance.size() == 0) {
        cout << "WARNING: getDecideNode(vector<unsigned> entry, TreeNode* node), empty entry\n";
        return nullptr;
    }
    
    if (node == NULL) {
        node = root;
    }
    
    if (node->isEndNode) {
        return node;
    }
    else {
        for (auto *child : node->children) {
            //cout << child->childIndex << " " << entry[node->value] << "\n";
            if (child->childIndex == instance[node->value]) {
                return getEndNode(instance, child);
            }
        }
    }
    cout << "ERROR: getDecideNode(vector<unsigned> entry, TreeNode* node), not finding entry in tree\n";
    return nullptr;
}

void EMTI_DecisionTree::mergeSubTree (EMTI_TreeNode* subTreeRoot , int insertPosition) {
    
    EMTI_TreeNode* subTreeRootCopy = new EMTI_TreeNode(*subTreeRoot);
    
    EMTI_TreeNode* oldNode = getSubTree(insertPosition);
    
    
    
    subTreeRootCopy->childIndex = oldNode->childIndex;
    subTreeRootCopy->setDepth(oldNode->depth);
    
    if (oldNode->parent != NULL) {
        
        EMTI_TreeNode* parent = oldNode->parent;
        
        subTreeRootCopy->parent = parent;
        
        parent->children[oldNode->childIndex] = subTreeRootCopy;
    }
    else {
        root = subTreeRootCopy;
    }
    
    
    delete oldNode;
    
    updateNodeCountAndList();
}

void EMTI_DecisionTree::changeNodeValue (int value, int position) {
    EMTI_TreeNode* node = getSubTree(position);
    
    node->value = value;
}


void EMTI_DecisionTree::updateNodeCountAndList () {
    
    pair<int,int> c = countNodes_andUpdateTree();
    
    decisionNodesCount = c.first;
    
    endNodesCount = c.second;
    
    nodesCount = decisionNodesCount + endNodesCount;
    
    
}

pair<int, int> EMTI_DecisionTree::countNodes_andUpdateTree (EMTI_TreeNode* node, int depth) {
    if (node == NULL) {
        node = root;
        nodeList.clear();
    }
    
    nodeList.push_back(node);
    node->depth = depth;
    
    if (node->isEndNode) {
        return pair<int,int>(0,1);
    }
    else {
        pair<int,int> subTreeSize = pair<int,int> (1,0);
        
        int childIndex = 0;
        for (auto &child : node->children) {
            child->childIndex = childIndex;
            subTreeSize = subTreeSize + countNodes_andUpdateTree(child, depth+1);
            childIndex++;
        }
        
        return subTreeSize;
    }
}

EMTI_TreeNode* EMTI_DecisionTree::getSubTree (int position) {
    if (position >= 0 && position < nodesCount) {
        return nodeList[position];
    }
    else {
        cout << "ERROR: getPointerOfPosition (int position), out of bounds position given to tree\n";
        return nullptr;
    }
}

EMTI_TreeIterator EMTI_DecisionTree::begin() {
    return EMTI_TreeIterator(root);
}

EMTI_TreeIterator EMTI_DecisionTree::end() {
    return EMTI_TreeIterator(nullptr);
}

string EMTI_DecisionTree::to_string() {
    
    string s = "root(d" +  ::to_string(root->depth) + ":" + ::to_string(root->value) + ") {";
    
    for (auto &child : root->children) {
        s += to_stringRecursive(child) + ",";
    }
    s.pop_back();
    s+= "}";
    
    return s;
}

string EMTI_DecisionTree::to_stringRecursive(EMTI_TreeNode* n) {
    
    if (n->isEndNode) {
        if (n->value == -1) return "end(d" +  ::to_string(n->depth) + ":X)";
        return "end(d" +  ::to_string(n->depth) + ":" + ::to_string(n->value) + ")";
    }
    else {
        
        string s = "node(d" +  ::to_string(n->depth) + ":" + ::to_string(n->value) + ") {";
        
        for (auto &child : n->children) {
            s += to_stringRecursive(child) + ",";
        }
        s.pop_back();
        s+= "}";
        
        return s;
    }
}

int EMTI_DecisionTree::random_number(int start, int end) {
    int range = end - start + 1;
    int number;
    
    if (end>start) {
        number = rand() % range + start;
    }
    else {
        number = start;
    }
    
    return number;
}



