//
//  EMTI_TreePruner.cpp
//  EMTI_0.2_swift
//
//  Created by Anders Nikolai Fuglseth on 18.04.2016.
//


#include "EMTI_TreePruner.hpp"

EMTI_DecisionTree*  EMTI_TreePruner::error_complexity_prune (EMTI_DecisionTree* tree, vector<vector<unsigned>>* trainingSet, vector<vector<unsigned>>* validationSet) {
    
    if (validationSet->size() == 0) return tree;
    
    vector<EMTI_DecisionTree*> candidates = getPrunedCandidates(tree, trainingSet);
    
    EMTI_DecisionTree* bestCandidate = candidates.front();
    double smallestError = 1;
    double standardError = 0;
    double chosenError = 1;
    
    for (EMTI_DecisionTree* candidate : candidates) {
        int wrong = 0;
        for (vector<unsigned> entry : *validationSet) {
            wrong += (candidate->decide(entry) != entry[entry.size()-1]);
        }
        double error = (double)wrong / validationSet->size();
        
        //cout << "Candidate error = " << error << " size = " << candidate->nodesCount << "\n";
        
        if (error <= smallestError) {
            smallestError = error;
            bestCandidate = candidate;
            
            standardError = sqrt((double)(error * (1 - error)) / validationSet->size());
            //cout << "  standardError = " << standardError << "\n";
        }
        else if (error < smallestError + standardError) {
            chosenError = error;
            bestCandidate = candidate;
        }
    }
    
    bestCandidate = new EMTI_DecisionTree(*bestCandidate);
    
    for (EMTI_DecisionTree* candidate : candidates) {
        delete candidate;
    }
    candidates.clear();
    
    
    
    return bestCandidate;
}

vector<EMTI_DecisionTree*> EMTI_TreePruner::getPrunedCandidates(EMTI_DecisionTree* tree, vector<vector<unsigned>>* trainingSet) {
    
    vector<EMTI_DecisionTree*> candidates;
    
    EMTI_DecisionTree* oldTree = new EMTI_DecisionTree(*tree);
    
    for (EMTI_TreeIterator i = oldTree->begin(); i != oldTree->end(); ++i) {
        i.node->clearEndNodeAccuracies();
        i.node->traffic = 0;
    }
    for (vector<unsigned> entry : *trainingSet) {
        
        if (entry.size()==(*trainingSet)[0].size()) {
            registerEntryUpToRoot(oldTree->getEndNode(entry), entry[entry.size()-1]);
        }
    }
    
    candidates.push_back(oldTree);
    
    while (candidates.back()->nodesCount > 1) {
        candidates.push_back(getNewNextCandidate(candidates.back(), trainingSet));
    }
    
    return candidates;
}

EMTI_DecisionTree* EMTI_TreePruner::getNewNextCandidate (EMTI_DecisionTree* tree, vector<vector<unsigned>>* trainingSet) {
    
    
    EMTI_DecisionTree* nextCandidate = new EMTI_DecisionTree(*tree);
    double smallestAlpha = numeric_limits<double>::max();
    EMTI_TreeNode* pruneNode = nextCandidate->root;
    
    for (EMTI_TreeIterator i = nextCandidate->begin(); i != nextCandidate->end(); ++i) {
        if (!i.node->isEndNode){
            int bestCategoryCount = 0;
            for (int i : i.node->endNodeAccuracies) {
                if (i > bestCategoryCount) {
                    bestCategoryCount = i;
                }
            }
            double node_errorCost = 1;
            if (i.node->traffic > 0) node_errorCost = ((double)(i.node->traffic-bestCategoryCount) / i.node->traffic) * ((double)i.node->traffic / trainingSet->size());
            
            
            double subTree_errorCost = 0;
            int subTreeLeaveCount = 0;
            for (EMTI_TreeNode* child : i.node->children) {
                subTree_errorCost += getSubTreeLeaves_errorCost(child, trainingSet->size());
                subTreeLeaveCount += getSubTreeLeaveCount(child);
            }
            
            double alpha = (node_errorCost - subTree_errorCost) / (subTreeLeaveCount - 1);
            if (alpha < smallestAlpha) {
                smallestAlpha = alpha;
                pruneNode = i.node;
            }
            if (alpha == smallestAlpha) {
                if (i.node->depth > pruneNode->depth) { //if all else equal, prune the deepest node first
                    pruneNode = i.node;
                }
            }
            
            //cout << i.node->childIndex << ":depth=" <<  i.node->depth << " alpha=" << alpha << " node_ErrorCost=" << node_errorCost << " subTree_errorCost=" << subTree_errorCost << " subTreeLeaveCount=" << subTreeLeaveCount << " " << bestCategoryCount << "/" << i.node->traffic << "\n";
        }
    }
    
    
    for (EMTI_TreeNode* child : pruneNode->children) {
        delete child;
    }
    pruneNode->children.clear();
    pruneNode->isEndNode = true;
    int bestCategoryCount = 0;
    for (int i = 0; i < pruneNode->endNodeAccuracies.size(); i++) {
        if (pruneNode->endNodeAccuracies[i] > bestCategoryCount) {
            bestCategoryCount = pruneNode->endNodeAccuracies[i];
            pruneNode->value = i;
        }
    }
    nextCandidate->updateNodeCountAndList();
    
    //cout << "-------------------------------\n";
    //    cout << "Before:" << tree->to_string() << "\n";
    //    cout << "After:" << nextCandidate->to_string() << "\n";
    
    return nextCandidate;
}

void EMTI_TreePruner::registerEntryUpToRoot(EMTI_TreeNode* n, unsigned value) {
    n->endNodeAccuracies[value]++;
    n->traffic++;
    if (n->parent != NULL) {
        registerEntryUpToRoot(n->parent, value);
    }
}

double EMTI_TreePruner::getSubTreeLeaves_errorCost (EMTI_TreeNode* n, long trainingSetSize) {
    if (n->isEndNode) {
        int bestCategoryCount = 0;
        for (int i : n->endNodeAccuracies) {
            if (i > bestCategoryCount) {
                bestCategoryCount = i;
            }
        }
        if (n->traffic == 0) return 1; //TODO find out whether to return 1 or 0 here
        return ((double)(n->traffic-bestCategoryCount) / n->traffic) * ((double)n->traffic / trainingSetSize);
    }
    else {
        double subTree_errorCost = 0;
        for (EMTI_TreeNode* child : n->children) {
            subTree_errorCost += getSubTreeLeaves_errorCost(child, trainingSetSize);
        }
        return subTree_errorCost;
    }
    return 0;
}

int EMTI_TreePruner::getSubTreeLeaveCount(EMTI_TreeNode* n) {
    if (n->isEndNode) {
        return 1;
    }
    else {
        int subTreeLeaveCount = 0;
        for (EMTI_TreeNode* child : n->children) {
            subTreeLeaveCount += getSubTreeLeaveCount(child);
        }
        return subTreeLeaveCount;
    }
    return 0;
}

EMTI_DecisionTree* EMTI_TreePruner::minimum_error_prune (EMTI_DecisionTree* tree, vector<vector<unsigned>>* trainingSet) {
    EMTI_DecisionTree* pruneTree = new EMTI_DecisionTree(*tree);
    
    for (EMTI_TreeIterator i = pruneTree->begin(); i != pruneTree->end(); ++i) {
        i.node->clearEndNodeAccuracies();
        i.node->traffic = 0;
    }
    for (vector<unsigned> entry : *trainingSet) {
        
        if (entry.size()==(*trainingSet)[0].size()) {
            registerEntryUpToRoot(pruneTree->getEndNode(entry), entry[entry.size()-1]);
        }
    }
    
    int bestCategoryCount = 0;
    for (int i : pruneTree->root->endNodeAccuracies) {
        if (i > bestCategoryCount) {
            bestCategoryCount = i;
        }
    }
    
    double root_prune_expected_error_rate = expected_error_rate((int)pruneTree->root->endNodeAccuracies.size(), pruneTree->root->traffic, bestCategoryCount);
    
    double root_noprune_expected_error_rate = 0;
    
    for (EMTI_TreeNode* child : pruneTree->root->children) {
        if (pruneTree->root->traffic == 0) {
            root_noprune_expected_error_rate += 1;
        }
        else {
            root_noprune_expected_error_rate += ((double)child->traffic / pruneTree->root->traffic) * do_prune_min_err(child);
        }
    }
    //cout << "prune_expected_error = " << root_prune_expected_error_rate << " vs noprune_expected_error = " << root_noprune_expected_error_rate << "\n";
    if (root_prune_expected_error_rate < root_noprune_expected_error_rate) {
        for (EMTI_TreeNode* child : pruneTree->root->children) {
            delete child;
        }
        pruneTree->root->children.clear();
        pruneTree->root->isEndNode = true;
        int bestCategoryCount = 0;
        for (int i = 0; i < pruneTree->root->endNodeAccuracies.size(); i++) {
            if (pruneTree->root->endNodeAccuracies[i] > bestCategoryCount) {
                bestCategoryCount = pruneTree->root->endNodeAccuracies[i];
                pruneTree->root->value = i;
            }
        }
    }
    pruneTree->updateNodeCountAndList();
    
    return pruneTree;
}

double EMTI_TreePruner::do_prune_min_err (EMTI_TreeNode* n) {
    int bestCategoryCount = 0;
    for (int i : n->endNodeAccuracies) {
        if (i > bestCategoryCount) {
            bestCategoryCount = i;
        }
    }
    
    double prune_expected_error_rate = expected_error_rate((int)n->endNodeAccuracies.size(), n->traffic, bestCategoryCount);
    
    if (n->isEndNode) {
        //cout << "    endNode:prune_expected_error = " << prune_expected_error_rate << "\n";
        return prune_expected_error_rate;
    }
    
    double noprune_expected_error_rate = 0;
    
    for (EMTI_TreeNode* child : n->children) {
        if (n->traffic == 0) {
            noprune_expected_error_rate += 0;
            //cout << "no traffic in this node\n";
        }
        else {
            noprune_expected_error_rate += (((double)child->traffic / n->traffic) * do_prune_min_err(child));
            //cout << "    decisionNode:noprune_expected_error_rate = " << noprune_expected_error_rate << "\n";
        }
    }
    //cout << "prune_expected_error = " << prune_expected_error_rate << " vs noprune_expected_error = " << noprune_expected_error_rate << "\n";
    
    if (prune_expected_error_rate < noprune_expected_error_rate) {
        //cout << "pruning!\n";
        for (EMTI_TreeNode* child : n->children) {
            delete child;
        }
        n->children.clear();
        n->isEndNode = true;
        int bestCategoryCount = 0;
        for (int i = 0; i < n->endNodeAccuracies.size(); i++) {
            if (n->endNodeAccuracies[i] > bestCategoryCount) {
                bestCategoryCount = n->endNodeAccuracies[i];
                n->value = i;
            }
        }
        return prune_expected_error_rate;
    }
    return noprune_expected_error_rate;

}

double EMTI_TreePruner::expected_error_rate (int categoryCount, int nodeTrafficCount, int bestCategoryCount) {
    
    return (double)(nodeTrafficCount - bestCategoryCount +  categoryCount - 1) / (nodeTrafficCount + categoryCount);
}















